/*

    File: main.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <scrnsave.h>
#include <gl/gl.h>
#include <math.h>
#include <stdio.h>

HDC hDC = NULL;
HGLRC hRC = NULL;
UINT timerID = 0;
int win_width = 0, win_height = 0;

bool using_window = false;

#define random() ((double)rand() / (double)RAND_MAX)

#define TWO_PI		(6.28318530717958647692528676655901)


char *dialogClassName = "TestProg";
char *windowName = "Within the Matrix SS - By Chris Moore";
int diag_winWidth, diag_winHeight, diag_winX, diag_winY;

double theta[3];
double time = 0.0;
double period[3], resettime = 0.0;
double color[3], nextcolor[3];

#define ROT_MAG		5.0

void InitRotate(void)
{
	theta[0] = theta[1] = theta[2] = 0.0;
	time = 0.0;

	for (int a = 0; a < 3; a++)
	{
		//color inits:
		color[a] = nextcolor[a];
		nextcolor[a] = random();
		//and of course - the rotation:
		period[a] = floor(random() * 30.0 + 10.0) * 25.0;
	}

	for (resettime = period[0];
		resettime< period[0] * period[1] * period[2];
		resettime += period[0])
	{
		if ((resettime / period[1]) == floor(resettime / period[1]))
		{
			if ((resettime / period[2]) == floor(resettime / period[2])) return;
		}
	}

}


void DoRotate(void)
{
	for (int a = 0; a < 3; a++)
	{
		//rotation
		theta[a] += ROT_MAG * sin(time * TWO_PI / period[a]);
	}
	time += 1.0;

	double d = time / resettime;
	glColor4d((nextcolor[0] - color[0]) * d + color[0],
				(nextcolor[1] - color[1]) * d + color[1],
				(nextcolor[2] - color[2]) * d + color[2], 0.25);

	if (time == resettime) InitRotate();

}

void DisplayScreen(void)
{
	glClear(GL_COLOR_BUFFER_BIT);

	DoRotate();

	glPushMatrix();

	glRotated(theta[0], 1.0, 0.0, 0.0);
	glRotated(theta[1], 0.0, 1.0, 0.0);
	glRotated(theta[2], 0.0, 0.0, 1.0);

	glTranslated(-10.0, -10.0, -10.0);

	for (int x = 0; x < 10; x++)
	{
		glPushMatrix();
		for (int y = 0; y < 10; y++)
		{
			glPushMatrix();
			for (int z = 0; z < 10; z++)
			{
				glBegin(GL_QUADS);

				glVertex3d(1.0, 1.0, 1.0);
				glVertex3d(1.0, -1.0, 1.0);
				glVertex3d(-1.0, -1.0, 1.0);
				glVertex3d(-1.0, 1.0, 1.0);

				glEnd();

				glTranslated(0.0, 0.0, 2.0);
			}
			glPopMatrix();
			glTranslated(0.0, 2.0, 0.0);
		}
		glPopMatrix();
		glTranslated(2.0, 0.0, 0.0);
	}

	glPopMatrix();

	glFlush();
	SwapBuffers(hDC);
}

void ResizeOpenGL(int x, int y)
{
	double ratio = (double)y / (double)x;
	glViewport(0, 0, (GLint)x, (GLint)y);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glFrustum(-1.0, 1.0, -ratio, ratio, 1.0, 10000.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void InitOpenGL(void)
{
	ResizeOpenGL(win_width, win_height);

	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	//alpha channel stuff:
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glEnable(GL_BLEND);
	glColor4d(0.0, 1.0, 0.0, 0.25);

	glTranslated(0.0, 0.0, -10.0);

	nextcolor[0] = 0.0;
	nextcolor[1] = 1.0;
	nextcolor[2] = 0.0;

	InitRotate();
}

int CALLBACK MyTimerProc(void)
{
	DisplayScreen();
	return 0;
}

void KillWindow(HWND hWnd)
{
	if (timerID) KillTimer(hWnd, timerID);

	wglMakeCurrent(NULL, NULL);
	ReleaseDC(hWnd, hDC);
	wglDeleteContext(hRC);
}

bool InitWindow(HWND hWnd)
{	
	PIXELFORMATDESCRIPTOR pfd;
	int pf;

	hDC = GetDC(hWnd);

	/* there is no guarantee that the contents of the stack that become
	   the pfd are zeroed, therefore _make sure_ to clear these bits. */
	memset(&pfd, 0, sizeof(pfd));
	pfd.nSize		= sizeof(pfd);
	pfd.nVersion	= 1;
	pfd.dwFlags		= PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	pfd.iPixelType	= PFD_TYPE_RGBA;
	pfd.cColorBits	= 32;

	pf = ChoosePixelFormat(hDC, &pfd);
	if (pf == 0) {
	MessageBox(NULL, "ChoosePixelFormat() failed:  "
		"Cannot find a suitable pixel format.", "Error", MB_OK); 
		return false;
	}
 
	if (SetPixelFormat(hDC, pf, &pfd) == FALSE) {
	MessageBox(NULL, "SetPixelFormat() failed:  "
		"Cannot set format specified.", "Error", MB_OK);
		return false;
	} 

	DescribePixelFormat(hDC, pf, sizeof(PIXELFORMATDESCRIPTOR), &pfd);

	hRC = wglCreateContext(hDC);
	wglMakeCurrent(hDC, hRC);

	/*	RECT rect;
	memset(&rect, 0, sizeof(rect));
	GetBoundsRect(hDC, &rect, 0);
	win_width = rect.left - rect.right; //GetSystemMetrics(SM_CXSCREEN);
	win_height = rect.bottom - rect.top; //GetSystemMetrics(SM_CYSCREEN);
	OutputMeSomeInfo(win_width, win_height);*/

	RECT winRect;
	GetClientRect(hWnd, &winRect);
	win_width = winRect.right;
	win_height = winRect.bottom;
/*
	if (using_window == false)
	{
		win_width = GetSystemMetrics(SM_CXSCREEN);
		win_height = GetSystemMetrics(SM_CYSCREEN);
	}
	else
	{
		win_width = diag_winWidth;
		win_height = diag_winHeight;
	}
*/

	//randomize via the time:
	SYSTEMTIME systime;
	GetSystemTime((LPSYSTEMTIME)&systime);
	srand(systime.wMinute +
		systime.wSecond +
		systime.wMilliseconds);

	InitOpenGL();

	if ((timerID = SetTimer(hWnd, 1, 5, (TIMERPROC)MyTimerProc)) == 0) return false;

	return true;
}

LRESULT WINAPI ScreenSaverProc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static PAINTSTRUCT ps;

	switch (message) {
  	case WM_ERASEBKGND:
		return 1;		//this should say "dont erase me"
	case WM_PAINT:
		BeginPaint(hWnd, &ps);
		EndPaint(hWnd, &ps);
		return 0;
	case WM_CREATE:
		using_window = false;
		if (InitWindow(hWnd) == false) PostQuitMessage(0);
		return 0;
	case WM_DESTROY:
		KillWindow(hWnd);
		return 0;
	}

	// Deal with any unprocessed messages. . .
	return DefScreenSaverProc(hWnd, message, wParam, lParam);
}

BOOL WINAPI ScreenSaverConfigureDialog (HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	return true;
}

LRESULT APIENTRY
DialogWndProc(
    HWND hWnd,
    UINT message,
    WPARAM wParam,
    LPARAM lParam)
{
	static PAINTSTRUCT ps;

  switch (message) {
  	case WM_ERASEBKGND:
		return 1;		//this should say "dont erase me"
	case WM_PAINT:
		BeginPaint(hWnd, &ps);
		EndPaint(hWnd, &ps);
		return 0;
	case WM_CREATE:
		if (InitWindow(hWnd) == false) PostQuitMessage(0);

		MessageBox(hWnd, "Within The Matrix\n"
					"A Screensaver demonstrating combinations\n"
					"of rotations along the three main axii.\n"
					"By Chris Moore\n"
					"Last Updated 9 March 2001",
					"About...",
					MB_OK);

		return 0;
	case WM_SIZE:
		ResizeOpenGL(LOWORD(lParam),HIWORD(lParam));
		PostMessage(hWnd, WM_PAINT, 0, 0);
		return 0;

    case WM_DESTROY:
        PostQuitMessage(0);
        return 0;
	case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_KEYDOWN:
//	case WM_KEYUP:
//	case WM_MOUSEMOVE:
		PostQuitMessage(0);
		return 0;

    }

    /* Deal with any unprocessed messages */
    return DefWindowProc(hWnd, message, wParam, lParam);
}


BOOL WINAPI RegisterDialogClasses (HANDLE hInst)
{
    WNDCLASS wndClass;
    HWND hWnd;
    MSG msg;

	using_window = true;

    // Define and register a window class
    wndClass.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
    wndClass.lpfnWndProc = DialogWndProc;
    wndClass.cbClsExtra = 0;
    wndClass.cbWndExtra = 0;
    wndClass.hInstance = (HINSTANCE)hInst;
    wndClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    wndClass.hCursor = LoadCursor(NULL, IDC_ARROW);
    wndClass.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
    wndClass.lpszMenuName = NULL;
    wndClass.lpszClassName = dialogClassName;
    RegisterClass(&wndClass);

    // Figure out a default size for the window
    diag_winWidth = GetSystemMetrics(SM_CYSCREEN) / 3;
    diag_winHeight = GetSystemMetrics(SM_CYSCREEN) / 3;

    // Create a window of the previously defined class
    hWnd = CreateWindow(
        dialogClassName,        // Window class's name
        windowName,             // Title bar text
        WS_OVERLAPPEDWINDOW |   // The window's style
        WS_CLIPCHILDREN |
        WS_CLIPSIBLINGS,
        diag_winX, diag_winY,             // Position
        diag_winWidth, diag_winHeight,    // Size
        NULL,                   // Parent window's handle
        NULL,                   // Menu handle
        (HINSTANCE)hInst,       // Instance handle
        NULL);                  // No additional data

    hDC = GetDC(hWnd);

    // Map the window to the screen
    ShowWindow(hWnd, SW_SHOW);

    // Message loop
    while (1) {
        // execute idle function while no messages to process
        if (GetMessage(&msg, NULL, 0, 0) != TRUE) {
            break;
        }
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    return msg.wParam;

}